#!/usr/bin/perl
#
my $revision = '$Id: Sanitizer.pm,v 1.32 2001/10/11 19:27:15 bre Exp $';
my $version = 'Anomy 0.0.0 : Sanitizer.pm';
#
##  Copyright (c) 2000 Bjarni R. Einarsson. All rights reserved.
##  This program is free software; you can redistribute it and/or modify
##  it under the terms of the GNU General Public License as published by
##  the Free Software Foundation; either version 2 of the License, or
##  (at your option) any later version.
#
##############################################################################
#
# NOTE:  Sanitizer development is for the most part sponsored by
#        FRISK Software International, http://www.frisk.is/.  Please
#        consider buying their anti-virus products to show your 
#        appreciation.
#
##############################################################################
#
# This is the email sanitizer engine.
#
# Most of the ideas in this script were borrowed from John D. Hardin's 
# "security through procmail" ruleset, which is available here:
# ftp://ftp.rubyriver.com/pub/jhardin/antispam/procmail-security.html
#
# Note that this script is a little differently licensed from the rest 
# of the Anomy tools because I borrowed GPL'd code from John's script.
#
# Documentation and new versions are here: http://mailtools.anomy.net/
#
#############################################################################
# Function naming conventions used in this file:
#
#   lowercase_names:   public object methods
#   BiCapitlizedNames: private object methods
#   CAPITALIZED_NAMES: functions (not assigned to an object)
#

##[ Package definition ]######################################################

package Anomy::Sanitizer;
use strict;
use Anomy::Log;
use Anomy::MIMEStream;
use Digest::MD5;
use IO::File;

BEGIN {
    use Exporter  ();
    use vars      qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);
    $VERSION      = do { my @r = (q$Revision: 1.32 $ =~ /\d+/g); sprintf "%d."."%02d" x $#r, @r };
    @ISA          = qw(Exporter);
    @EXPORT       = qw( );
    @EXPORT_OK    = qw( );
};

##[ Default configuration ]###################################################

my $default_config = 
{
    # Features.
    # Disable stuff by replacing 1s with 0s.
    #
    "feat_verbose"    => 1,  # Warn user about unscanned parts and Other Stuff.
    "feat_log_inline" => 1,  # Inline logs: 0 = Off, 1 =  Maybe, 2 = Force
    "feat_log_stderr" => 1,  # Print log to stderr.
	"feat_log_xml"    => 0,  # Attachment logs: XML format.
    "feat_log_trace"  => 0,  # Attachment logs: Complete (all trace info).
    "feat_files"      => 1,  # Enable filename-based policy decisions.
    "feat_force_name" => 0,  # Force all parts (except text/plain parts) to
                             # have file names.
    "feat_boundaries" => 0,  # Replace all boundary strings with our own.
                             # NOTE:  Always breaks PGP/MIME messages!
    "feat_lengths"    => 1,  # Protect against buffer overflows.
    "feat_scripts"    => 1,  # Defang incoming shell scripts.
    "feat_html"       => 1,  # Defang active HTML content.
    "feat_trust_pgp"  => 0,  # Trust PGP signed messages -> don't scan them
    "feat_uuencoded"  => 1,  # Sanitized UU encoded attachments.
    "feat_forwards"   => 1,  # Sanitize forwarded messages.
    "feat_testing"    => 0,  # Enable to turn off randomness, for testcases
    "feat_fixmime"    => 1,  # Try and fix invalid MIME.
    "feat_paranoid"   => 0,  # Be very paranoid about MIME headers etc.

    ##########################################################################
    # If feat_files is non-zero, the following rules will be used to decide
    # what to do with an attachment.  The rules are all filename based, each
    # "list" being a regulaur expression.
    #
    # The file is compared to each list in order (1 to file_list_max) and on
    # the first match the defined policy is enforced.  If a file matches no
    # lists the default policy is used.
    #
    # Valid policies are:
    #
    #   mangle  - Completely ofbuscates the file name.
    #   defang  - Defangs the file name, without making it completely 
    #             illegible.
    #   accept  - Attachment is accepted as-is (possibly subject to 
    #             HTML or shell script defanging though).
    #   save    - Save the attachment to the "file_save_dir" directory,
    #             replace it with an informative message.
    #   drop    - Delete the attachment
    #   unknown - Indeterminate result, check the next policy.
    #   panic   - Return immediately with a non-zero exit code.
    #
    # In addition, if an exclamation mark (!) is appended to a policy, then
    # in addition to enforcing the policy, the internal bug score will be
    # incremented past the "score_bad" value, causing the sanitizer to 
    # return a non-zero exit code whnt it is finished.
    #
    # If a policy has four values, e.g. "save:save!:drop:save", then the file 
    # will be scanned for viruses using an external virus scanner.  Which of 
    # the four policies is used then depends on whether the result is "clean"
    # (1st), "successfully disinfected" (2nd), "unsuccessfully disinfected" 
    # (3rd) or "scan failed" (4th).
    #
    # The scanner definitions are as follows:
    #
    #   "e1:e2:e3:/path/to/scanner args ... %FILENAME ..."
    #
    # The e1, e2, e3 are comma-delimited lists of exit codes that match
    # the four different "interesting" return values we exped scanners to 
    # return.  Unexpected values are assumed to be in the "scan failed" 
    # category.
    #
	# In addition to the %FILENAME variable, the following variables are
	# also expanded: 
	#
	#    %REPLY_TO        The apparent reply-to address for the message.
	#    %ERRORS_TO       The apparent error-address for the message.
	#    %HEADER(<name>)  The named header of the top level message.  The
	#                     name must be in lowercase, ex: %HEADER(subject).
   	#

    "file_list_rules" => 3,

    # This is the file name template, used for creating (temporary?) files
    # when scanning or saving attachments.  The following substitutions are
    # supported:
    #               $d - Day of month (01-31)
    #               $m - Month number (01-12)
    #               $y - Two digit year (00-99)
    #               $Y - Four digit year
    #               $H - Hour (00-23)
    #               $M - Minute (00-59)
    #               $S - Second (00-59)
    #
    #               $P - This process's PID, in hex.
    #               $T - The current Unix time, in hex.
    #               $F - A safe version of the original file name.
    #               $  - A random character, from [A-Z0-9].
    #
    # It's recommended that all file name templates contain a few '$'
    # characters, since a new name will be generated (up to five times, 
    # after that it will give up) if the chosen one is already in use.  
    # More '$' substitions will mean fewer collisions.  Note that any
    # directories must exist, the sanitizer will NOT create them for you.
    # So if you are using random directory hashing make sure to create 
    # all the directories ahead of time!
    #
    "file_name_tpl" => '/tmp/att-$F-$T.$$',

    # This defines file names which always get mangled.
    # FIXME:  This list is very incomplete!
    "file_list_1_scanner" => 0,
    "file_list_1_policy" => "mangle",
    "file_list_1" => '(?i)'.
        '((happy99|x-mas|setup|aol4free|ie0199).exe|winmail.dat)',

    # This defines file names which don't get mangled, unless they are on
    # the blacklist.  Anything not on this list gets mangled.
    #
    "file_list_2_scanner" => 0,
    "file_list_2_policy" => "accept",
    "file_list_2" => '(?i)\.'.
        '(gif|tiff?|jpe?g|pn[mg]|x[pb]m|dvi|e?ps|p(df|cx)'. # Graphics
        '|mp[32]|wav|au|ram?'.                              # Sound
        '|avi|mov|mpe?g'.                                   # Movies
        '|z(ip|oo)|ar[cj]|lha|[tr]ar|rpm|deb|slp|tgz'.      # Archives
        '|t(xt|ex)|csv|l(og|yx)'.                           # Text
        '|[ch](pp|\+\+)?|s|inc|asm|pa(tch|s)|java|php\d?'.  # Uncompiled code
        '|[ja]sp'.                                          # ...
    #
    # The following depends on the HTML sanitizer doing it's job properly.
    #
        '|[sp]?html?'.                                      # HTML
        ')(\.g?z|\.bz2?)*$',                                # Compressed?

    # Scan attachments that look like they're Microsoft documents for
    # viruses, using John D. Hardin's macro scanning code.  A score of 
    # 25 or above is considered "infected" and the file is quarantined.
    #
    "file_list_3_scanner" => '0:1:2:builtin/macro 25',
    "file_list_3_policy" => "accept:save:save:defang",
    "file_list_3"        => '(?i)\.(do[tc]|xl[sw]|p[po]t|rtf)$',

    # Define some empty lists, so people don't have to edit the source to
    # define policies with more lists.
    "file_list_4_policy"  => 0,
    "file_list_4_scanner" => 0,
    "file_list_4"         => 0,
    "file_list_5_policy"  => 0,
    "file_list_5_scanner" => 0,
    "file_list_5"         => 0,
    "file_list_6_policy"  => 0,
    "file_list_6_scanner" => 0,
    "file_list_6"         => 0,
    "file_list_7_policy"  => 0,
    "file_list_7_scanner" => 0,
    "file_list_7"         => 0,
    "file_list_8_policy"  => 0,
    "file_list_8_scanner" => 0,
    "file_list_8"         => 0,

    # This defines the default policy, for filenames that don't match
    # any of the preceding lists.
    "file_default_policy" => "defang",
    
    # The default name for files of an unrecognized MIME-type which lack
    # a file-name.  The default is to treat such files as text files, 
    # both for backwards compatibility and to increase the chances that
    # unnamed attachments get treated "safely" by the recpient's MUA.
    "file_default_filename" => "unnamed.txt",

    # Characters permitted in file names.  In addition to these, the
    # characters "." and "-" are always allowed.  Set to 0 to allow
    # all characters.
    "file_characters" => 'A-Za-z0-9_=',

    # HTML tag blacklist, the following tags are defanged.  If the "STYLE"
    # tag is on the list, then style= entries within other tags will also
    # be defanged.
    "html_evil_tags" => 'META|LINK|FORM|APP|SCRIPT|OBJECT|EMBED|I?FRAME|FRAMESET|I?LAYER|STYLE',

    # Javascript defanging.  If non-zero, javascript, cid and about URLs 
    # will also be defanged.
    "html_javascript" => 'On(Abort|Blur|Change|Click|DblClick|DragDrop|Error|Focus|Key(Down|Press|Up)|Load|Mouse(Down|Move|Out|Over|Up)|Move|Re(set|size)|Select|Submit|Unload)',

    # MIME attributes for inline sanitizer logs.
    "sanitizer_log_disp" => 'attachment; filename="sanitizer.log"',
    "sanitizer_log_type" => 'text/sanitizer-log; charset="iso-8859-1"',

    ##########################################################################
	# Headers to add if they are currently missing, since some email 
	# clients expect them to be present.  This only applies to the top-level
	# message headers.

    "force_headers"   => 0,
	"force_header_1"  => "MIME-Version: 1.0",
	"force_header_2"  => "Content-Type: text/plain; charset=\"%DEF_CHARSET\"",
	"force_header_3"  => "Subject: (no subject)",
	"force_header_4"  => 0,
	"force_header_5"  => 0,
	"force_header_6"  => 0,
	"force_header_7"  => 0,
	"force_header_8"  => 0,
	"force_header_9"  => 0,
	"force_header_10" => 0,

    ##########################################################################

    # Messages.  Translate?
    #
    "header_info" => "X-Sanitizer: This message has been sanitized!",
    "header_url"  => "X-Sanitizer-URL: http://mailtools.anomy.net/",
    "header_rev"  => "X-Sanitizer-Rev: $revision",

    "var_def_charset" => "iso-8859-1",

    "msg_file_drop" => 
        "****\nNOTE:  An attachment was deleted from this part of the message,\n".
        "because it failed one or more checks by the virus scanning system.\n".
        "See the attached sanitization log for more details or contact your\n".
        "system administrator.\n\n".
        "The removed attachment's name was:\n\n".
        "\t%FILENAME\n\n".
        "It might be a good idea to contact the sender and warn them that\n".
        "their system is infected.\n****\n",

    "msg_file_save" => 
        "****\nNOTE:  An attachment was deleted from this part of the message,\n".
        "because it failed one or more checks by the virus scanning system.\n".
        "The file has been quarantined on the mail server, with the following\n".
        "file name:\n\n".
        "\t%SAVEDNAME\n\n".
        "The removed attachment's original name was:\n\n".
        "\t%FILENAME\n\n".
        "It is recommended that you contact your system administrator if you\n".
        "need access to the file.  It might also be a good idea to contact the\n".
        "sender, and warn them that their system may be infected.\n****\n",

    "msg_pgp_warning" => 
        "WARNING:  The following data has NOT been sanitized, to ensure\n".
        "          that the signature remains intact, if valid.  Please\n".
        "          be careful if you open any enclosed attachments.\n\n",

    "msg_log_prefix" =>
        "This message has been 'sanitized'.  This means that potentially\n".
        "dangerous content has been rewritten or removed.  The following\n".
        "log describes which actions were taken.\n",

    "msg_usage" =>
        "$version\n$revision\n\n".
        "Usage: sanitizer.pl [ 'variable op value' | 'filename' ] ... \n".
        "\n".
        "FIXME:  unwritten\n",
    
    "msg_defanged" => "DEFANGED",
    "msg_blacklisted" => "BLACKLISTED",
    "msg_panic" => "PANIC",

    "msg_current" =>
        "Current configuration:\n",

    "msg_signature" => "$version\n$revision\n",

    # Limits
    "max_conf_recursion" => 5,
};

my $score_config =
{
    ##########################################################################
    # Scoring system.
    # Use 0 to disable each threshold.
    #
    "score_bad_code" => 1,   # The "bad" exit code returned by the Sanitizer.
    "score_bad"      => 100, # Any message exceeding this value will cause
                             # the sanitizer to return a non-zero exit code
                             # after processing the entire message.

    "score_panic_code" => 2, # The "panic" exit code returned by the Sanitizer.
    "score_panic"      => 0, # If the sanitizer's internal score exceeds this
                             # value, the sanitizer will terminate immediately
                             # with a non-zero exit code.
};

my @default_configs = 
(
    $default_config,
	$score_config,
);

# Default file-names for unnamed MIME parts.
#
my $default_filenames = 
{
    "text/plain" => "unnamed.txt",
    "text/html"  => "unnamed.html",
    "image/gif"  => "unnamed.gif",
    "image/jpeg" => "unnamed.jpg",
    "image/png"  => "unnamed.png",
    "image/tiff" => "unnamed.tiff",
    "inline/text/plain"   => "unnamed.txt",
    "inline/text/html"    => "unnamed.html",
    "application/ms-tnef" => "winmail.dat",
};

# Default parser hash
my $default_parsers =
{
    # Search text parts for inline uuencoded attachments, so we can mangle
	# their names and possibly scan the attachments themselves.  This also
	# takes care of defanging HTML.
	"text/*" => \&CleanText,
	"application/pgp" => \&CleanText,
	"inline/text/*" => \&CleanText,

    # Check headers, parse contents...
	"multipart/*" => \&CleanMultipart,

    # Sanitize encapsulated messages.
	"message/rfc822" => \&CleanRFC822,
	"inline/message/rfc822" => \&CleanRFC822,

    # We don't recognize this content-type, see if we can figure anything
	# out from the headers themselves.  Sanitize headers (at least).
	"DEFAULT" => \&CleanUnknown,
};

# Built in scanners
my $default_scanners = 
{
};


##[ Public interface routines ]###############################################

sub new
{
    my ($proto, $name) = @_;
	my $class = ref($proto) || $proto;

	my $self = {
		# Configuration
		"conf"      => { },
		"defaults"  => [ @default_configs ],
		"parsers"   => { },
		"filenames" => { },
		"scanners"  => { },

		# Object globals
	    "log"         => new Anomy::Log,
		"mod_id"      => 0,
		"base_mod_id" => 0,
		"logd_mod_id" => 0,
		
		# Common part data, exposed.
		"common"      => undef,
		
		# Errors
		"error" => undef,
	};
	bless ($self, $class);

	$self->reset_config();
	%{ $self->{"parsers"} }   = %{ $default_parsers };
	%{ $self->{"filenames"} } = %{ $default_filenames };
	%{ $self->{"scanners"} }  = %{ $default_scanners };

	return $self;
}

sub error
{
    my $self = shift;
	return $self->{"error"};
}

sub reset_config
{
    my $self = shift;

    foreach my $c (@{ $self->{"defaults"} })
    {
        foreach my $v (keys(%{ $c }))
	    {
	        $self->{"conf"}->{$v} = $c->{$v};
	    }
    }
}

# Parses a single line from a configuration file.
# Configuration files look like this:
#
#   # this is a comment
#   # set some variables
#   variable = value
#   variable += value
#   # load another configuration file
#   /path/to/another/configuration/file
# OR:
#   include /path/to/another/configuration/file
#
# All white space in the value is replaced with spaces.   A " #" sequence
# (white space followed by '#') marks the beginning of a comment, and is
# ignored.  
#
# The following escape sequences are expanded in the value string to let
# you get around these "features":
#
#       \\  ->  \
#       \#  ->  #
#       \n  ->  newline
#       \t  ->  tab
#       \s  ->  space
#
# Using the .= or += instead of just = will append the string to the
# variable, instead of resetting it.
#
my $config_recursion = 0;
sub configure
{
    my $self = shift;
	my $conf = $self->{"conf"};

    while (my $line = shift)
    {
        if ($line =~ /^\s*([a-z0-9_]+)\s*([\+\.]*=)\s*(.*)\s*$/si)
		{
            # OK, this lookes like a variable configuration
            my ($var, $op, $val) = (lc($1), $2, $3);

            unless (($var =~ /^var_/i) || (defined $conf->{$var})) 
            {
                $self->{"error"} .= "Unknown configuration variable: $var\n";
				return 1;
            }

            $val =~ s/\s+#.*$//;
            $val =~ tr/\t\n/  /;
            $val =~ s/\\([\\nts#])/ { $_=$1; tr|\\nts#|\\\n\t #|; $_ } /eg;

            if ($op eq '=')
            {           
                $conf->{$var} = $val;
            }
            else
            {
                $conf->{$var} .= $val;
            }
        }
        elsif ($line !~ /^\s*(#.*)?$/)
        {
            # Ooh, we're supposed to include another configuration file!

            my $fn = $line;
			$fn =~ s/^\s*include\s+//i;
            $fn =~ s/^\s*(.*)\s*$/$1/;

            $config_recursion++;
			if ($config_recursion > $conf->{"max_conf_recursion"})
			{
			    $self->{"error"} .= "Configuration files nested too deep!\n";
				return 2;
			}

            local *CF;
			unless ((-r $fn) && (open (CF, "< $fn")))
			{
			    $self->{"error"} .= "Cannot read $fn: $!\n";
				return 3;
			}

            my $ln = 0;
            while (my $cl = <CF>)
            {
                $ln++;
                chomp $cl;
                if ($self->configure($cl))
                {
                    $config_recursion--;
					$self->{"error"} .= "[$config_recursion] Error in $fn, line $ln.\n";
					return 4;
                }
            }
            $config_recursion--;
        }
    }
	
	if (0 == $config_recursion)
	{
	    # Side effects of configuration...
		if ($conf->{"feat_trust_pgp"})
		{
            # Only scan headers of signed stuff, if we decide that we
			# trust messages that are signed or encrypted.
			#
			$self->{"parsers"}->{"multipart/signed"} = \&CleanHeaders;
			$self->{"parsers"}->{"multipart/encrypted"} = \&CleanHeaders;
		}

        # FIXME: Should move into sanitizer.pl!
		# Sanity check for scoring system
		#
		if (($conf->{"score_panic"}) && 
            ($conf->{"score_panic"} < $conf->{"score_bad"}))
        {
            $conf->{"score_bad"} = $conf->{"score_panic"};
		}
	}

    return undef;
}

sub get_config_text
{
    my $self = shift;
	my $conf = $self->{"conf"};

    my $cfg = "";
    for my $key (sort(keys(%{ $conf })))
    {
        my $val = $conf->{$key};
		$key = sprintf("%-20s", $key);
    
        $val =~ s/\n/\\n\n$key += $1/g;
		$val =~ s/\t/\\t/g;
		$val =~ s/\#/\\#/g;
		$val =~ s/\\\\/\\/g;
#       $val =~ s/(.{55})/$1/g;
        $val =~ s/\+=  /+= \\s/g;
		$val =~ s/ \n/\\s\n/g;
		$val =~ s/\n\S+\s+[\+\.]?=\s*$//ms;
    
	    $cfg .= $key ."  = ". $val ."\n";
    }
    return $cfg;
}

sub get_msg
{
    my ($self, $var) = @_;
    return $self->{"conf"}->{"msg_".lc($var)};
}
sub expand_msg
{
    my ($self, $msg) = @_;
    return $self->expand($self->get_msg($msg));
}

sub get_var
{
    my ($self, $var) = @_;
    return $self->{"conf"}->{"var_".lc($var)};
}
sub expand_var
{
    my ($self, $var) = @_;
    return $self->expand($self->get_var($var));
}
sub set_var
{
    my ($self, $var, $value) = @_;
    $self->{"conf"}->{"var_".lc($var)} = $value;
}

sub expand
{
    my ($self, $msg) = @_;
    $msg =~ s/%([A-Z_]+)/ $_=($self->{"conf"}->{"var_".lc($1)} || "%".$1) /ge;
    return $msg;
}

sub sanitize
{
    my ($self, $fh_in, $fh_out) = @_;
	my $conf = $self->{"conf"};

    # Create MIMEStream object
    my $message = Anomy::MIMEStream->New($fh_in, $fh_out, 
	                                     $self->WrappedParsers());

    # Record the common data for use by hooks.
    $self->{"common"} = $message->{"common"};

	# Register this run with the core logger.
	my $start = time();
	$start = 0 if ($conf->{"feat_testing"});
	$self->{"log"}->sublog("Sanitizer", SLOG_TRACE, 
	                       { start => $start }, $message->{"log"});

    # Parse message header
    $message->ParseHeader();

    # Log that we've parsed the header - this creates an event allowing
	# hooks to reconfigure most of the sanitizer based on header data.
	# WARNING: MIMEStream parsers are NOT reconfigurable at this point.
    $message->{"log"}->entry("parsed_header", SLOG_TRACE, undef,
	                         "Finished parsing message header.");

    # Initialize modification ID counter.
    $self->{"mod_id"} = ($$ * 1001) % (10 ** (rand() * 6));

    # Disable randomness when testing
    $self->{"mod_id"} = 99 if ($conf->{"feat_testing"});
	
    # Store initial value for comparisons later.
    $self->{"base_mod_id"} = $self->{"mod_id"};
    $self->{"logd_mod_id"} = $self->{"mod_id"};

    # Append blurb to header.
    {
        my $t;
		chomp $message->{"rawheader"};
		$message->{"rawheader"} .= $self->expand("$t\n") if ($t = $conf->{"header_info"});
		$message->{"rawheader"} .= $self->expand("$t\n") if ($t = $conf->{"header_url"});
		$message->{"rawheader"} .= $self->expand("$t\n") if ($t = $conf->{"header_rev"});
		
		for my $n (1..$conf->{"force_headers"})
		{
		    my $h = $conf->{"force_header_$n"};
		    next unless $h;
		    my ($hdr, $val) = split(/:\s+/, $h);
			unless ($message->{"headers"}->{lc($hdr)})
			{
			    $message->{"rawheader"} .= $self->expand("$h\n");
				$message->{"log"}->entry("force_header", SLOG_WARNING, 
				                         { header => $hdr, value => $val },
				                         "Added default value (%value%) for header %header%");
			}
		}
		
		$message->{"rawheader"} .= "\n";
    }

	# Force message to be multipart/mixed, if it isn't already that or
	# plain text which we know we can append a log to.
	if (($conf->{"feat_log_inline"} > 1) && 
	    ($message->{"mime"}->{"type"} !~ /^(multipart\/mixed|text\/(plain|html))$/i))
	{
	    $message->{"log"}->entry("feat_log_inline", SLOG_WARNING, undef,
		                         "Forcing message to be multipart/mixed, to facilitate logging.");

        my $mt = lc($message->{"mime"}->{"type"});
        $message->{"parsers"}->{"ORG/$mt"} = $message->{"parsers"}->{"$mt"};
		$message->{"parsers"}->{"$mt"} = sub { return $self->WrapWithMultipart(@_); };
	}

    # Parse everything else!
	$message->ParseBody();

    # Dump log to STDERR if requested, things like that.
    $self->DumpLog($self->{"log"}, undef, 1);
	
    return 0;
}

sub register_scanner
{
    my $self = shift;
    my $name = shift;
    my $scanner = shift;

    $self->{"scanners"}->{lc($name)} = $scanner;
}

##[ Helper routines ]##########################################################

sub WrappedParsers
{
    my $self = shift;
	my $wrapped = { };
	
	foreach my $p (keys(%{ $self->{"parsers"} }))
	{
	    my $c = $self->{"parsers"}->{$p}; 
	    $wrapped->{$p} = sub { return &$c($self, @_); };
	}

	return $wrapped;
}


##[ Sanitizers, output, etc. ]#################################################


# This will print out the contents of $log, if the time looks right.
#
# Unfortunately, we can't guarantee that the time will /ever/ be right,
# so this may fail to embed the logs in the message - they are guaranteed
# to go to stderr though.  The logs try to be as unobtrusive as possible.
#
sub DumpLog
{
    my $self = shift;
    my $plog = shift;
	my $conf = $self->{"conf"};
	
    my $part = shift;
    my $finished = shift;
    my $type = $part->{"mime"}->{"type"} if ($part);

    my $inline = (($conf->{"feat_log_inline"}) && ($part));

    my $alevel = SLOG_WARNING|SLOG_ERROR;
    $alevel = SLOG_ALL if ($conf->{"feat_log_trace"});

	my $astext = $self->{"log"}->print_as_text($alevel);
	$inline = 0 if ($astext =~ /^\s*$/);

    my $ppart;
    if (($inline) && ($ppart = $part->{"parent"}))
    {
        $inline = 0 if ($ppart->{"parent"});
        $inline = 0 if ($ppart->{"mime"}->{"type"} =~ /^(multipart\/mixed|text\/)/i);
        $inline = 0 if ($type !~ /^text\/(plain|html)/i);
		$inline = 0 if ($part->{"sanitizer_dumped_log"});
    }

    my $prelog = \$conf->{"msg_log_prefix"};
    my $signature = "";
    $signature = $conf->{"msg_signature"} unless ($conf->{"feat_testing"});

    if (($inline || $finished) && 
        ($self->{"logd_mod_id"} != $self->{"mod_id"}))
	{
	    # Record total changes (so far).
		$self->{"logd_mod_id"} = $self->{"mod_id"};
		$plog->entry("modifications", SLOG_INFO, 
		             { base  => $self->{"base_mod_id"},
					   end   => $self->{"mod_id"}, 
					   total => $self->{"mod_id"}-$self->{"base_mod_id"} },
					 "Total modifications so far: %total%");
	}

    if ($inline)
    {
	    $alevel |= SLOG_INFO;
        if ($type =~ /^multipart\/mixed/i) 
        {
            $part->Write(
                "--". $part->{"mime"}->{"boundary"} ."\n".
                "Content-Type: ". $conf->{"sanitizer_log_type"} ."\n".
                "Content-Transfer-Encoding: 8bit\n".
                "Content-Disposition: ". $conf->{"sanitizer_log_disp"} ."\n\n");
			if ($conf->{"feat_log_xml"})
			{
                $part->Write("<Sanitizer_Log>\n <blurb>\n".$self->expand($$prelog)."\n </blurb>\n");
			    $part->Write($self->{"log"}->print_as_xml($alevel));
				$part->Write($self->expand(" <RevTxt>\n$signature\n </RevTxt>\n</Sanitizer_Log>\n"));
			}
			else
			{
                $part->Write($self->expand($$prelog)."\n");
			    $part->Write($self->{"log"}->print_as_text($alevel));
				$part->Write($self->expand("\n$signature\n"));
			}
			$part->{"sanitizer_dumped_log"} = 1;
        }
		elsif ($conf->{"feat_log_inline"} > 1)
		{
		    # Do nothing, we know a multipart opportunity will show
			# up eventually. :-)
		}
        elsif ($type =~ /^(text\/plain|application\/pgp)/i)
        {
            $part->Write($self->expand("\n-- \n$$prelog\n"));
			$part->Write($self->{"log"}->print_as_text($alevel));
			$part->Write($self->expand("\n$signature\n"));
			$part->{"sanitizer_dumped_log"} = 1;
        }
        elsif ($type =~ /^text\/html/i) 
        {
            $part->Write($self->expand("<BR><HR><TABLE BORDER=1 BGCOLOR=\"white\"><TR><TD><B>$$prelog</B><P>\n"));
			$part->Write($self->{"log"}->print_as_html($alevel, "", "black"));
			$part->Write($self->expand("<P>$signature<P></TD></TR></TABLE>\n"));
			$part->{"sanitizer_dumped_log"} = 1;
        }
    }
    if (($finished) && ($conf->{"feat_log_stderr"}))
    {
	    $alevel |= SLOG_INFO;
	    if ($conf->{"feat_log_xml"})
		{
		    print STDERR $self->{"log"}->print_as_xml($alevel);
		}
		else
		{
            print STDERR $self->{"log"}->print_as_text($alevel);
		}
    }
}


# This will create a new file based on the "file_name_tpl" template.
# Returns undef on failure.
#
my $caf_inc = 0;
my @charray = ('A'..'Z', 0..9);
sub CreateAttFile
{
    my $self = shift;
	my $conf = $self->{"conf"};
    my $fn = shift;
    my $ofn = shift;
    my $fh = undef;
    my $cnt = 0;

    # This is the file name we use on disk - keep it as simple as possible.
    $ofn = Anomy::MIMEStream::Encode7bit(undef, $ofn);
    $ofn =~ s/[^A-Za-z0-9\.-]/_/gs;

    do
    {
        my $T = time();
        my ($S, $M, $H, $d, $m, $y) = localtime($T);
        $$fn = $conf->{"file_name_tpl"} || return undef;

        # Date stuff
        $$fn =~ s/\$T/ sprintf("%x", $T) /eg;
        $$fn =~ s/\$S/ sprintf("%2.2d", $S) /eg;
        $$fn =~ s/\$M/ sprintf("%2.2d", $M) /eg;
        $$fn =~ s/\$H/ sprintf("%2.2d", $H) /eg;
        $$fn =~ s/\$d/ sprintf("%2.2d", $d) /eg;
        $$fn =~ s/\$m/ sprintf("%2.2d", $m) /eg;
        $$fn =~ s/\$y/ sprintf("%2.2d", $y % 100) /eg;
        $$fn =~ s/\$Y/ sprintf("%d", $y + 1900) /eg;

        # PID
        $$fn =~ s/\$P/ sprintf("%x", $$) /eg;
        
        # Safe file name
        $$fn =~ s/\$F/$ofn/g;

        # Random characters
        if ($conf->{"feat_testing"})
        {
            $$fn =~ s/\$/ $charray[ ($caf_inc++ % 35) ] /eg;
        }
        else
        {
            $$fn =~ s/\$/ $charray[ int(rand(35.99)) ] /eg;
        }
    }
    while (($cnt++ < 5) && 
           (!defined ($fh = IO::File->new($$fn, O_CREAT|O_EXCL|O_RDWR))));

    return $fh;
}


# Scan a file for viruses, using the given command string.
#
sub ScanFile
{
    my $self = shift;
    my $plog = shift;
    my ($e1, $e2, $e3, $cmd) = split(/:/, shift, 4);
    my $filename = shift;
    my $fh = shift;
    my $md5x2 = shift;  # This is a double MD5 sum of the attachment data.

    return 0 unless (defined $cmd);

    my $log = $plog->sublog("ScanFile", SLOG_TRACE, { file => $filename });

    my @args = map {
	                 $_ =~ s/%FILENAME/$filename/gi; 
	                 $_ =~ s/%REPLY_TO/$self->{common}->{"reply-to"}/gi;
	                 $_ =~ s/%ERRORS_TO/$self->{common}->{"errors-to"}/gi;
	                 $_ =~ s/%HEADER\(([a-z_-]+)\)/$self->{common}->{headers}->{$1}/gi;
					 $_ 
				   } split(/\s+/, $cmd);
    my $spid = undef;
    my $sleeps = 0;
	my $result = -9999;

    $log->entry("command", SLOG_TRACE|SLOG_DEBUG, undef, $cmd);

    # Use built-in scanner if requested.
    if (lc($args[0]) =~ /^builtin\/?(.*)$/)
    {
        my $cmd = $self->{"scanners"}->{lc($1)};
		my $junk = shift @args;

		$result = (256 * &$cmd($self, $log, $fh, $md5x2, @args)) 
		  if (defined $cmd);
    }
	else
	{
        # Flush buffers, before forking.
		STDOUT->flush();
		STDERR->flush();

        do
		{
            unless (defined ($spid = open(SCANNER, "-|")))
			{
                $log->entry("error", SLOG_ERROR, { text => "$!" }, 
			                "Cannot fork: %text%");

                return 3 if ($sleeps++ > 6);
                sleep(10);
            }
        } until (defined $spid);

        if (!$spid) # Are we the kid?
		{
            print STDOUT "Scan cmd: ", join(' ', @args), "\n";
		
		    # We want the scanner's stderr to be sent to stdout!
            close(STDERR) && 
              open(STDERR, ">&STDOUT") || 
                print STDOUT "WARNING:  Couldn't dup STDOUT!\n";

            STDOUT->flush();

            unless (exec { $args[0] } @args)
            {
                print STDOUT "Exec failed: $!\n";
                die "Exec failed: $!";
            }
        }

        # Not the kid, read the scanner's output.
        while (my $l = <SCANNER>)
        {
            $log->entry("output", SLOG_TRACE|SLOG_DEBUG, undef, $l);
        }
        close (SCANNER);
		$result = $?;
    }

    # Was file clean?
    for my $v (split(/,/, $e1))
    {
        if ($result == (256 * $v))
        {
			$log->entry("result", SLOG_INFO, 
	                    { summary => "clean", code => $result }, 
     	                "Scan succeeded, file is clean.");
            return 0;
        }
    }

    # Was file dirty, but is now clean?
    for my $v (split(/,/, $e2))
    {
        if ($result == (256 * $v))
        {
	        $log->entry("result", SLOG_WARNING|SLOG_INFO, 
                        { summary => "disinfected", code => $result }, 
                        "File was infected, but the virus checker fixed it.");
            return 1;
        }
    }

    # Was file dirty and unfixable?
    for my $v (split(/,/, $e3))
    {
        if ($result == (256 * $v))
        {
	        $log->entry("result", SLOG_WARNING|SLOG_INFO, 
	                    { summary => "infected", code => $result }, 
                        "File was infected, the virus checker couldn't fixed it.");
            return 2;
        }
    }

    $log->entry("result", SLOG_WARNING|SLOG_INFO|SLOG_ERROR, 
                { summary => "error", code => $result },
      	        "Unknown exit code: %code%");

    return 3;
}


# Clean/scan a file, sanitize the file name.  
#
# This will change the part's reader pointer to point to a virtual on-disk
# part, if a virus scanner is used.
#
sub SanitizeFile
{
    my ($self, $part, $mime) = @_;
	my $conf = $self->{"conf"};

    my $unknown = undef;
    my $fnp = \$unknown;
    my $typep = \$unknown;
    my $sinc = 0;
    
    for my $fhn ("name", "filename")
    {
        $fnp = \$mime->{$fhn} if ($mime->{$fhn}); 
    }
    $typep = \$mime->{"type"} if ($mime->{"type"}); 

    # Don't apply rules to multipart or message parts, or inline
    # parts (e.g. forwarded messages).
    return undef if ($$typep =~ /^m(ultipart|essage)\//i);

    my $log = new Anomy::Log;
    
    # If no file name is specified, create one from the MIME type.
    if (!defined $$fnp) 
    {
        my $t = $default_filenames->{lc($$typep)};
        $t = $conf->{"file_default_filename"} if (!defined $t);
        $fnp = \$t if ($t);

        # Abort if the part still has no file name.
        return undef unless ($$fnp);

        # Forcibly add a file name to unnamed parts?
        if (($conf->{"feat_force_name"}) &&
            ($part->{"mime"}->{"type"}) && 
            ($part->{"mime"}->{"type"} !~ /^(inline\/)?text\/plain/i))
        {
            $part->{"mime-headers"}->{"content-type"} .= " name";
            $mime->{"name"} = $t;
            $fnp = \$mime->{"name"};
            
            $log->entry("default_name", SLOG_INFO, { default => $t },
                        "No attachment name found, using default (%default%).");
        }
    }

    # Insert our log into part log.
    $part->{"log"}->sublog("SanitizeFile", SLOG_TRACE,
                           { filename => $$fnp, mimetype => $$typep }, $log);

    # Store original file name & type
    my $ofn = $$fnp;
    my $otype = $$typep;
    my $pol = undef;
    my @matched = ( );
    my $fh = undef;
    my $filename = undef;
    my $scanned = 0;

    for my $i (1..$conf->{"file_list_rules"}, "default")
    {
        $log->entry("Check_Rule", SLOG_TRACE, 
		            { rule => $i, file => $$fnp, list => $conf->{"file_list_$i"} },
					"Rule %rule%: %file% =~ %list%")
          if ($i ne "default");

        if (($i eq "default") || ($$fnp =~ $conf->{"file_list_$i"}))
        {
            my $mlog = $log->sublog("Match", SLOG_INFO, { rule => $i });
			
            push @matched, $i;
            my @policy = split(":", $conf->{"file_default_policy"});

            @policy = split(":", $conf->{"file_list_${i}_policy"})
                if ($conf->{"file_list_${i}_policy"});

            my $scan_result = 0;
            if ((@policy == 1) && (lc($policy[0]) =~ /^drop\!?$\s*/))
            {
                while ($part->Read()) { };
            }
            elsif ((@policy == 4) || (lc($policy[0]) =~ /^save\!?\s*$/))
            {
                # Create a file name from our template.
                my $ofh = $fh;
                if (($fh) || ($fh = $self->CreateAttFile(\$filename, $$fnp)))
                {
				    my $md5_1 = new Digest::MD5;
					my $md5_2 = new Digest::MD5; $md5_2->add("Shift!");
					my $size = 0;

                    if (!$ofh)
                    {
                        # Save the attachment to disk...
                        while (my $l = $part->Read())
                        {
                            # FIXME:  need better error handling!
                            $fh->print($l) || die;
							$md5_1->add($l);
							$md5_2->add($l);
							$size += length($l);
                        }

                        $fh->flush();
                    }
					
					# Log stuff.
					my $dig = $md5_1->hexdigest().$md5_2->hexdigest();
					$mlog->entry("saved-file", SLOG_TRACE, 
					             { file => $filename, digest => $dig, size => $size },
								 "Saved attachment as %file% (%size% bytes, digest %digest%).");

                    # We want to scan this attachment
                    if (@policy == 4)
                    {
                        $scan_result = 
                            $self->ScanFile($mlog,
							                $conf->{"file_list_${i}_scanner"},
											$filename, $fh, $dig);
                        $scanned = 1;
                    }
                }
                else
                {
                    # Error :(	
  	                $mlog->entry("error", SLOG_WARNING|SLOG_INFO|SLOG_ERROR, undef, 
		                         "Failed to create temporary file for scanning attachment!");
                    $scan_result = 3;
                }
            }
            $pol = lc($policy[$scan_result]) || "defang";
            $pol =~ s/\s*$//g;

            $sinc += $conf->{"score_bad"} if ($pol =~ s/\!$//);

			my $llev = SLOG_INFO;
			$llev |= SLOG_WARNING if ($pol !~ /accept/i);
	        $mlog->entry("policy", $llev, { name => $pol }, 
	                     "Enforced policy: %name%");

            # Enforce policy, based on scan result.
            if ($pol eq "mangle")
            {
                $sinc++;
                $$fnp = $conf->{"msg_blacklisted"} .".". $conf->{"msg_defanged"} ."-".$self->{"mod_id"};
                $$typep = "application/". $conf->{"msg_defanged"} ."-".$self->{"mod_id"};
            }
            elsif ($pol eq "defang")
            {
                $sinc++;
                $$fnp =~ s/\./_/g;
                $$fnp .= ".". $conf->{"msg_defanged"} ."-".$self->{"mod_id"};
                $$typep = "application/". $conf->{"msg_defanged"} ."-".$self->{"mod_id"};
            }
            elsif ($pol =~ /^(drop|save)$/i)
            {
                $sinc++;

                my $what = $1;
                my $fn = $filename;
                $fn =~ s/^.*\///g;

	            $conf->{"var_filename"} = $$fnp;
				$conf->{"var_savedname"} = $fn;

                my $msg = $self->expand_msg("file_". lc($what));

                $part->UnRead($msg);

                # Fix encoding etc.
                $mime->{"encoding"} = "8bit";
                $mime->{"charset"} = $conf->{"var_def_charset"};
                $mime->{"disposition"} = "inline";
                $part->{"encoder"} = $part->{"encoders"}->{"8bit"};
                if ($part->{"uupart"})
                {
                    $part->{"postponed"} = undef;
                    $part->{"rawheader"} = undef;
                    $part->{"uupart"} = 0;
                }
                $$typep = "text/plain";
                $$fnp = $conf->{"msg_defanged"} ."-".$self->{"mod_id"}.".txt";

				# Make sure our modifications are actually used!
				$part->{"mime-headers"}->{"content-type"} = "type charset";
				$part->{"mime-headers"}->{"content-type"} .= " name" if ($mime->{"name"});
				$part->{"mime-headers"}->{"content-encoding"} = "type charset";

                # This keeps the file from getting reincluded.
                $fh = undef;

                # This may keep the file from getting deleted.
                $filename = undef if ($pol eq "save");
            }
            elsif ($pol eq "panic")
            {
                $$fnp .= "-" . $conf->{"msg_panic"};
                if ($conf->{"score_panic"})
                {
                    $sinc += $conf->{"score_panic"};
                }
                else
                {
                    $conf->{"score_panic"} = 10000;
                    $sinc += 10000;
                }
            }

            if (defined $fh)
            {
                # Prepare to re-read attachment from disk.
                # I suppose this is bad OO-form.
                $fh->seek(0, SEEK_SET);
                my $reader = Anomy::MIMEStream->New($fh);
                $part->{"reader"} = $reader;
                $reader->{"writer"} = $part;
            }

            # Keep trying?
            next if ($pol eq "unknown");

            # Cleanup.
            unlink $filename if ($filename);

            last;
        }
    }

    # Remove spooky characters from file name.
	if ($conf->{"file_characters"})
	{
        $$fnp =~ s/[^$conf->{"file_characters"}\.-]/_/g;
	}

    # Truncate file name length, by chopping off anything preceding the
    # last 80 characters - the extension matters more than the beginning
    # of the name.
    $sinc++ if ($$fnp =~ s/^.*(.{80})$/$1/);

    # Format the list of rules matched.
    my $matches = join(", ", @matched);
    $matches =~ s/-1/default/;

    if (($ofn) && ($$fnp ne $ofn))
    {
		$sinc++;

        # Mangle x-mac-* headers if present, so Eudora won't work 
        # around the mangling.
        if ($part->{"rawheader"} =~ s/^x-mac/X-$conf->{"msg_defanged"}\[$self->{"mod_id"}\]-MAC/gim)
		{
	        $log->entry("defanged-x-mac", SLOG_WARNING|SLOG_INFO, 
		                { id => $self->{"mod_id"} }, 
	                    "Defanged part's X-Mac headers.");
		}

        # Log change...
	    $log->entry("new-mimetype", SLOG_WARNING|SLOG_INFO, 
		            { id => $self->{"mod_id"}, value => $$typep }, 
	                "Replaced mime type with: %value%")
             if (($otype) && ($otype ne $$typep));
        $log->entry("new-filename", SLOG_WARNING|SLOG_INFO, 
		            { id => $self->{"mod_id"}, value => $$fnp },
	                "Replaced file name with: %value%");

    }

    if ($mime->{"filename"})
    {
        $mime->{"name"} = $mime->{"filename"};
    }
	
	# Increment modification ID.
	$self->{"mod_id"}++ if ($sinc);
}


# Truncate just about anything, printing out a blurb at the same
# time.  The fourth argument is a replacement value to use instead
# of the overly-long original one.  Omitting it just truncates the
# field.
#
sub Truncate
{
    my $self = shift;
    my $dataname = shift;
    my $data = shift;
    my $maxlen = shift;
    my $safeval = shift;
    my $log = shift;

    if ($maxlen < length($$data))
    {
        $safeval = substr($$data, 0, $maxlen) unless ($safeval);

        $log->entry("truncated", SLOG_WARNING|SLOG_INFO, 
	                { id   => $self->{"mod_id"}, 
					  what => $dataname,
					  old  => $$data,
					  new  => $safeval,
					  max  => $maxlen },
					"Rewrote long ( >%max% bytes ) %what% from:\n".
					"    >>%old%<<\n".
					" to >>%new%<<");
					
        $self->{"mod_id"}++;
        $$data = $safeval;
    }
}


# This is a helper function for SanitizeHTML
#
sub SanitizeTag
{
    my $self = shift;
	my $conf = $self->{"conf"};

    my $tag = shift;
	my $log = shift;
    my $otag = $tag;

    # Don't defang email addresses!
    return $tag if ($tag =~ /<\S+\@/);

    # Unquote things that have no good reason to be quoted - potentially
    # dangerous obfuscated URLs, for the most part.
    if ($tag =~ /["'][^"'\s]*&#x?[1-9][0-9a-f]/si) 
    {
        while ($tag =~ /["'][^"'\s]*&#((4[6-9]|5[0-8]|6[4-9]|[78][0-9]|9[07-9]|1[0-1][0-9]|12[0-2]))/s) 
        {
            my $char = chr($1);
            $tag =~ s/&#$1;?/$char/sg;
        }
        while ($tag =~ /["'][^"'\s]*&#(x(2[ef]|3[0-9a]|4[0-9a-f]|5[0-9a]|6[1-9a-f]|7[0-9a]))/si) 
        {
            my $char = chr(hex("0$1"));
            $tag =~ s/&#$1;?/$char/sgi;
        }
    }
    if ($tag =~ /["'][^"'\s]*%[2-7][0-9a-f]/si) {
        while ($tag =~ /["'][^"'\s]*%((2[ef]|3[0-9a]|4[0-9a-f]|5[0-9a]|6[1-9a-f]|7[0-9a]))/si)
        {
            my $char = chr(hex("0x$1"));
            $tag =~ s/%$1/$char/sgi;
        }
    }

    # Defang dangerous tags & attributes
    $tag =~ s/(<(?:!--\s*)?)($conf->{"html_evil_tags"})/${1}$conf->{"msg_defanged"}_$2/sgi;
    if ($conf->{"html_evil_tags"} =~ /STYLE/i)
    {
        $tag =~ s/\s(STYLE\s*)=/ $conf->{"msg_defanged"}_$1=/sgi;
    }
    $tag =~ s/$conf->{"html_javascript"}/$conf->{"msg_defanged"}_On$1/sgi;

    # Defang javascript: and cid: urls.
    # Should I add SMB: URLs to this list?
    if ($conf->{"html_javascript"})
    {
        $tag =~ s/(["'])(about|cid|[a-z]+script|mocha):/${1}$conf->{"msg_defanged"}_$2:/sgi;
        $tag =~ s/(["'])&{/${1}$conf->{"msg_defanged"}_&{/sg;
    }

    # Kill various types of web-bugs
    if (($conf->{"html_webbugs"}) && 
        ($tag !~ /<I?FRAME/i))
    {
        $tag =~ s/(src=["']?)(http|ftp|mailto|smb)/$1$conf->{"msg_defanged"}_$2/isg;
    }
    
    # Logging
    if ($otag ne $tag)
    {
        my $ntag = $tag;

        $ntag =~ s/\s+/ /gs;
        $ntag =~ s/[<>]/_/gs;
        $otag =~ s/\s+/ /gs;
        $otag =~ s/[<>]/_/gs;
		$log->entry("rewrote-html", SLOG_WARNING|SLOG_INFO, 
		            { id => $self->{"mod_id"}, old => $otag, new => $ntag }, 
					"Rewrote HTML tag: >>%old%<<\n".
					"              as: >>%new%<<");
		$self->{"mod_id"}++;
    }

    return $tag;
} 

# This routine sanitizes a snippet of HTML.
#
# If the tag is incomplete, it is returned as a "leftover" for later
# scanning, else the function returns undef.  The calling function is
# expected to detect overly long leftovers, and truncate them if 
# necessary.
#
# This is based on John's HTML sanitization code.  I may replace this 
# later with a more complete parser which defangs any attributes not 
# specifically white-listed.
#
sub SanitizeHTML
{
    my $self = shift;
	my $conf = $self->{"conf"};

    my $lp = shift;
	my $log = shift;
    my $leftovers = undef;

    # Check for trailing tags...
    if ($$lp =~ s/(<(?:!--\s+)?[A-Za-z]+[^>]+)$//s)
    {
        $leftovers = $1;
        if (length($leftovers) > 2048) 
        {
            my $lo = $leftovers;
            
            $lo =~ s/[<>]/_/gs;
            $lo =~ s/\s+/ /gs;

			$log->entry("split-html", SLOG_WARNING|SLOG_INFO, 
			            { id => $self->{"mod_id"},
						  begin => substr($lo, 0, 25), 
						  end => substr($lo, -25, -1) }, 
						"Split really long tag (over 2k):\n".
						"    >>%begin% ... %end%<<");

            $$lp .= $leftovers . '>';
            $leftovers = "<". $conf->{"msg_defanged"} .".".$self->{"mod_id"}." ";
            $self->{"mod_id"}++;
        }
    }

    # Defang active HTML content
    $$lp =~ s/(<[A-Za-z]+.*?>)/ $self->SanitizeTag($1, $log) /gse;

    return $leftovers;
}

# This routine will truncate and rewrite message headers, to block 
# buffer-overflow exploits and filename-based trojans.
#
sub SanitizeHeaders
{
    my $self = shift;
	my $conf = $self->{"conf"};

    my $part = shift;
    my $boundfix = shift;
    my $headers = $part->{"headers"};
    my $mime = { };
    
    # Copy headers...
    foreach my $key (keys(%{ $part->{"mime"} }))
    {
        $mime->{$key} = $part->{"mime"}->{$key};
    }

    my $old_mod_id = $self->{"mod_id"};
	my $log = $part->{"log"};

    # Header length checks, more strict than the generic tests below.
    if ($conf->{"feat_lengths"})
    {
        $self->Truncate("MIME content-type", \$mime->{"type"},     80, "application/octet-stream", $log);
        $self->Truncate("MIME charset",      \$mime->{"charset"},  40, $conf->{"var_def_charset"}, $log);
        $self->Truncate("MIME encoding",     \$mime->{"encoding"}, 40, "8bit", $log);

        # This is more strict than the following header scan, and is 
        # designed to explicitly defang the Outlook Date: overflow.
        if ($part->{"rawheader"} =~ s/(Date:\s*[^\n]{40,40})([^\n]+)/$1\nX-$conf->{"msg_defanged"}-Date: [$self->{"mod_id"}] $2/gs)
        {
            $log->entry("split-date", SLOG_WARNING|SLOG_INFO, 
			            { id => $self->{"mod_id"} },
						"Split unusually long Date: header.");
			$self->{"mod_id"}++;
        }
        # This will limit the length of each individual word in the headers
        # to 128 characters, inserting a text marker and spaces when longer
        # words are encountered.  This is designed to foil attacks based on 
        # vulnerabilities like those described in various bugtraq posts 
        # in July 2000, including the USSR-2000050 advisory.
        #
        my $spaces = " " x (1 + rand()*7);
        $spaces = " " if ($conf->{"feat_testing"});
        if ($part->{"rawheader"} =~ s/(\S{128,128})/$conf->{"msg_defanged"}\[$self->{"mod_id"}]:$1$2 "$spaces" /gs)
        {
            $log->entry("split-words", SLOG_WARNING|SLOG_INFO, 
			            { id => $self->{"mod_id"} },
                        "Split unusually long word(s) in header.");
			$self->{"mod_id"}++;
        }
        # Make sure subject line is of a reasonable length, if it looks 
        # remotely like it contains a file name.
        if (($headers->{"subject"} =~ /\.\S+/) &&
            (length($headers->{"subject"}) > 128))
        {
            my $s = $headers->{"subject"};
            $s =~ s/^.*(.{100,100})$/$1/s;
            $part->{"rawheader"} =~ s/(Subject:.*?)\Q$headers->{"subject"}\E/$1$conf->{"msg_defanged"}\[$self->{"mod_id"}]:$s/si;
            
            $s = $headers->{"subject"};
            chomp $s;

            $log->entry("truncated-subject", SLOG_WARNING|SLOG_INFO, 
			            { id => $self->{"mod_id"}, value => $s },
						"Truncated long subject line:\n    >>%value%<<");
			$self->{"mod_id"}++;
        }
    }

    # File name sanity checks...
    if ($conf->{"feat_files"})
    {
        $self->SanitizeFile($part, $mime);
    }
    
    # Replace the boundary string: we know ours is sane! :-)
    if ((($conf->{"feat_boundaries"}) || 
         ($mime->{"boundary"} eq "") ||
         (($conf->{"feat_fixmime"}) && ($mime->{"undecoded-boundary"} =~ /[\(\)]/))) && 
        ($boundfix) && (defined $mime->{"boundary"}))
    {
        $part->{"bad-mime-boundary"} = $mime->{"boundary"};
        if ($conf->{"feat_testing"})
        {
            $mime->{"boundary"} = "MIMEStream=_+testing".$self->{"mod_id"};
        }
        else
        {
            $mime->{"boundary"} = Anomy::MIMEStream::MakeBoundary();
        }

        $log->entry("replaced-boundary", SLOG_WARNING|SLOG_INFO, 
	                { id  => $self->{"mod_id"},
					  old => $part->{"bad-mime-boundary"},
					  new => $mime->{"boundary"} },
					"Replaced MIME boundary: >>%old%<<\n".
					"                  with: >>%new%<<");
		$self->{"mod_id"}++;
    }

    # Fix encoding of output so it conforms to MIME standard.
    if (($conf->{"feat_fixmime"}) && 
        (!$part->{"uupart"}) &&
        ($mime->{"type"} =~ m"^m(ultipart|essage)/"i) &&
        ($mime->{"encoding"} !~ /^([78]bit|binary)$/i))
    {
        if (my $e = $part->{"encoders"}->{ "8bit" })
        {
            $log->entry("fixed-encoding", SLOG_WARNING|SLOG_INFO, 
			            { id => $self->{"mod_id"} },
                        "Fixed invalid part encoding.");
			$self->{"mod_id"}++;
            $mime->{"encoding"} = "8bit";
            $part->{"encoder"} = $e;
        }
        else
        {
            $log->entry("error-encoding", SLOG_WARNING|SLOG_INFO|SLOG_ERROR, undef,
                        "Couldn't fix part invalid encoding!");
        }
    }

    # Rebuild the MIME headers with sane/safe values.
    my $newheaders = "";
    for my $header ("Content-Type",
                    "Content-Transfer-Encoding",
                    "Content-Disposition")
    {
        next unless (defined $headers->{lc($header)});

        my @fields = split(/\s+/, $part->{"mime-headers"}->{lc($header)} );
        my $t = undef;

        for my $field (@fields)
        {
            my $value = $mime->{lc($field)};

            $self->Truncate("MIME $field", \$value, 100, $log) 
			  if ($conf->{"feat_lengths"});

            my $oval = $value;
            if ($value =~ /^\s*$/)
            {
				$log->entry("dropped-mime", SLOG_WARNING|SLOG_INFO, 
				            { id => $self->{"mod_id"},
							  field => $field },
							"Dropped empty MIME field: %field%");
				$self->{"mod_id"}++;
            }
            else
            {
                if (($conf->{"feat_paranoid"}) && ($value =~ s/[\'\"\`\$]/_/g))
                {
					$log->entry("rewrote-mime", SLOG_WARNING|SLOG_INFO, 
				                { id    => $self->{"mod_id"},
								  field => $field, 
								  old   => $oval,
								  new   => $value },
								"Rewrote MIME field %field% as\n".
								"    >>%new%<< (was >>%old%<<)");
					$self->{"mod_id"}++;
                }
                
                if ($t)
                {
                    $t .= "; $field=\"". $part->EncodeHeader($value) ."\"";
                }
                else
                {
                    $t = $value;
                }
            }
        }
        $newheaders .= $header .": ". $t ."\n";

        # Modifying the $headers directly here should be okay, since we are
        # inserting equivalent values (not identical) unless new bugs are 
        # found, in which case the raw headers will be updated.
        $headers->{lc($header)} = $t;
    }

    # Only modify part header if absolutely necessary.
    my $writer = $part->{"writer"} || $part;
    if (($conf->{"feat_boundaries"}) || ($self->{"mod_id"} > $old_mod_id))
    {
        $writer = Anomy::MIMEStream->Writer($part, $mime);
        $writer->KillRawMimeHeaders();
        $writer->{"rawheader"} =~ s/\s*$//s;
        $writer->{"rawheader"} .= "\n" if ($writer->{"rawheader"} ne "");
        if (!$writer->{"parent"})
        {
            $writer->{"rawheader"} .= "MIME-Version: 1.0\n";
            $headers->{"mime-version"} = "1.0\n";
        }
        $writer->{"rawheader"} .= $newheaders . "\n";
    }

    # Remove any active HTML from header.
    if ($conf->{"feat_html"})
    {
        my $rhl = $self->SanitizeHTML(\$writer->{"rawheader"}, $log);
        $writer->{"rawheader"} .= $rhl;
    }
}


##[ Parsers ]##################################################################

# Wrap a part in a multipart/mixed wrapper.
#
sub WrapWithMultipart
{
    my $self = shift;
	my $conf = $self->{"conf"};

    my $part = shift;
    my $reader = $part;
    $part->{"log"}->entry("parser", SLOG_TRACE, undef, "WrapWithMultipart");

	# Create a new boundary string
	my $bound = Anomy::MIMEStream::MakeBoundary();
	$bound = "Test_Wrapper_Boundary" if ($conf->{"feat_testing"});

	# Update headers
	my $old = $part->KillRawMimeHeaders();
	
    # Buffer hacks.  This is terribly ugly code.
 	$part->{"EncodeBase64"} = undef;
	my $buf = "--".$bound.$old."\n\n";
	$buf .= &{ $part->{"encoder"} }($part, $part->{"IOBuffer"});
	$buf .= &{ $part->{"encoder"} }($part, undef);
	# Hack to handle internal state in Base64 decoder.
	if ($part->{"DecodeBase64"})
	{
	    $buf .= $part->{"DecodeBase64"};
	}
	$part->{"DecodeBase64"} = undef;
	$part->{"IOBuffer"} = $buf;

    # Fix parser table
	my $mt = lc($part->{"mime"}->{"type"});	
	$part->{"parsers"}->{"$mt"} = $part->{"parsers"}->{"ORG/$mt"};

    # Create writer
    my $w = Anomy::MIMEStream->Writer($part, 
	        { 
			    "boundary" => $bound, 
				"type"     => "multipart/mixed", 
				"encoding" => "8bit",
			},{
			    "content-type"              => "type boundary",
			    "content-transfer-encoding" => "encoding",
			});

	# Header hack, to save a few cycles.
	$part->{"mime"} = $w->{"mime"};
	$part->{"mime-headers"} = $w->{"mime-headers"};
	undef $part->{"mime-headers"}->{"content-disposition"};

	# Log hack
	$part->{"log"} = $w->{"log"};

    # Encoding hack.
	$w->{"EncodeBase64"} = $part->{"EncodeBase64"} = undef;
	$w->{"DecodeBase64"} = $part->{"DecodeBase64"} = undef;
	$part->{"decoder"} = \&Anomy::MIMEStream::Decode8bit;
	$part->{"encoder"} = \&Anomy::MIMEStream::Encode8bit;

	return $self->CleanMultipart($part);
}

# This sanitizes a text part, and it's headers.
#
sub CleanText
{
    my $self = shift;
	my $conf = $self->{"conf"};

    my $part = shift;
    my $reader = $part;
    $part->{"log"}->entry("parser", SLOG_TRACE, undef, "CleanText");
    
    my $ishtml = 0;
    if (($part->{"mime"}->{"filename"} =~ /html?$/i) ||
        ($part->{"mime"}->{"name"} =~ /html?$/i) ||
        ($part->{"mime"}->{"type"} =~ /html?$/i))
    {
        $ishtml = 1;
    }

    my $in_trusted_text = 0;
    $in_trusted_text = 100
        if (($conf->{"feat_trust_pgp"}) &&
            ($part->{"mime"}->{"type"} =~ /multipart\/signed/i));

    $self->SanitizeHeaders($part, 1);
    my $writer = $part->{"writer"} || $part;
    $writer->WriteHeader();

    # Some sanitizations only make sense at the very top of a file. 
    # This determines how many non-blank lines are "near the top".
    my $neartop = 5;

    my $leftovers = undef;
    while (my $l = $reader->Read())
    {
        $l = $leftovers . $l if ($leftovers);

        if ($l =~ /^-+BEGIN.*?SIGNED\s+MESSAGE-+\s*$/smi)
        {
            if ($conf->{"feat_trust_pgp"})
            {
                $in_trusted_text++;

				$part->{"log"}->entry("trust-signed", SLOG_INFO,
				                      { id => $self->{"mod_id"}, 
									    silently => $conf->{"feat_verbose"} },
			                 "Disabled scanning for signed part of message.");

                if ($conf->{"feat_verbose"})
                {
                    $writer->Write($self->expand_msg("pgp_warning"));
                }
				$self->{"mod_id"}++;
            }
        }

        if ($in_trusted_text)
        {
            $in_trusted_text-- if ($l =~ /^-+END.*?SIGNATURE-+\s*$/smi);
        }
        # This "else" is safe, since a PGP boundary is harmless.
        else
        {
            # Check for inline forwarded messages.
            if (($conf->{"feat_forwards"}) && 
                ($l =~ s/^(---+.*?Forward.*?---+\s*)$//smi))
            {
                my $fwd = $1;

                # Deal with leftover html snippets by closing them.
                if (($conf->{"feat_html"}) && ($l !~ /^\s*$/))
                {
                    $l .= $conf->{"msg_defanged"} .".".$self->{"mod_id"}.">\n";
					$part->{"log"}->entry("closed-html", SLOG_WARNING|SLOG_INFO, 
					                      { id => $self->{"mod_id"} },
			                              "Closed open HTML tag preceding forwarded message.");
                    $self->{"mod_id"}++;
                    $self->SanitizeHTML(\$l, $part->{"log"});
                    $writer->Write($l);
                }

                $writer->Write($fwd);
                $l = $reader->ParserForwardedMessage();
            }
            # Check for inline uuencoded attachments.  
            # Sanitize their contents.
            if (($conf->{"feat_uuencoded"}) && 
                ($l =~ s/^(begin \d\d\d\d? \S+.*)$//smi))
            {
                $reader->UnRead($1);
                
                # Deal with leftover html snippets by closing them.
                if (($conf->{"feat_html"}) && ($l !~ /^\s*$/))
                {
                    $l .= $conf->{"msg_defanged"} .".".$self->{"mod_id"}.">\n";
					$part->{"log"}->entry("closed-html", SLOG_WARNING|SLOG_INFO,
					                      { id => $self->{"mod_id"} },
			                              "Closed open HTML tag preceding uuencoded attachment.");
					$self->{"mod_id"}++;
                    $self->SanitizeHTML(\$l, $part->{"log"});
                    $writer->Write($l);
                }

                $l = Anomy::MIMEStream::ParserUUAttachment($reader);
            }
            elsif ($neartop)
            {
                # Check for Unix shell scripts.
                if (($conf->{"feat_scripts"}) && 
                    ($l =~ s/^#!/#!\/bin\/sh\necho $conf->{"msg_defanged"}.$self->{"mod_id"}\nexit\n#!/gsm))
                {
					$part->{"log"}->entry("defanged-shell", SLOG_WARNING|SLOG_INFO,
					                      { id => $self->{"mod_id"} },
			                              "Defanged UNIX shell script(s).");
					$self->{"mod_id"}++;
                }
            }

            # Sanitize embedded HTML - we do this last so our hiding stuff
            # in $leftovers won't disable the other checks.
            $leftovers = $self->SanitizeHTML(\$l, $part->{"log"}) 
			    if ($conf->{"feat_html"});
        }

        $neartop-- if (($neartop) && ($l !~ /^\s*$/));

        if ((!$in_trusted_text) && ($ishtml) && 
            ($l =~ s/^(.*)(<\/(?:BODY|HTML))/$2/si))
        {
            $writer->Write($1);
            $self->DumpLog($part->{"log"}, $writer);
        }
        $writer->Write($l);
    }
    $writer->Write($leftovers) if ($leftovers);

    # Append log "signature style"
    $self->DumpLog($part->{"log"}, $writer) if (!$in_trusted_text);

    # Flush buffers
    $writer->Write(undef);
}


# This sanitizes the headers of an otherwise unfamiliar part.
#
sub CleanUnknown
{
    my $self = shift;
    my $part = shift;
    my $guess = $part->GuessMimeType();
    $part->{"log"}->entry("parser", SLOG_TRACE, undef, "CleanUnknown");

    if ($guess =~ /text\//i)
    {
        # Why wasn't this marked as text/plain or text/html ?
        return $self->CleanText($part);
    }

    return $self->CleanHeaders($part);
}


# This sanitizes only the headers of a part.
#
sub CleanHeaders
{
    my $self = shift;
    my $part = shift;
    
    $self->SanitizeHeaders($part, 0);
    Anomy::MIMEStream::ParserCat($part);

    # BUG: Can't dump logs here, could royally screw up message.
}


# This sanitizes the headers of a message/rfc822 part, before processing
# the part itself with the default handlers.
#
sub CleanRFC822
{
    my $self = shift;
    my $part = shift;
    $part->{"log"}->entry("parser", SLOG_TRACE, undef, "CleanRFC822");

    $self->SanitizeHeaders($part, 0);
    Anomy::MIMEStream::ParserRFC822($part);

    # BUG: Can't dump logs here, could royally screw up message.
}


# This sanitizes the headers of a multipart part, before processing the
# part itself with the default handlers.
#
sub CleanMultipart
{
    my $self = shift;
    my $reader = shift;
    $reader->{"log"}->entry("parser", SLOG_TRACE, undef, "CleanMultipart");

    $self->SanitizeHeaders($reader, 1);
    Anomy::MIMEStream::ParserUnclosedMultipart($reader);
    my $writer = $reader->{"writer"} || $reader;

    # Append sanitization log as seperate text part, if this is the end 
    # of the outermost multipart/* part.
    #
    $self->DumpLog($writer->{"log"}, $writer);

    # Garbage collection & cleanup.
    $writer->Write(undef);
    $writer->Close();
    $writer->Amputate();
}

#EOF#
1;
